!+
! Module: epexposure_mod
!
! Authors:  Michael Freyberg - XMM-SSC-MPE
!
! Description: spatial exposure corrections for
!              for EPIC pn scientific observations
!
! Comments: 
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!-
module epexposure_mod

  use dal
  use cal
  use oal
  use param
  use errorHandling

  implicit none

  private

! Global variables

  public epexposure

contains

  !---------------------------------------------------------------------
  subroutine epexposure
  !---------------------------------------------------------------------

  character(len=10), parameter :: CodeVersion = '0.9'

  type(DataSetT)                :: infile
  type(TableT)                  :: expo, dlimap, events  ! , inccd
  type(ColumnT)                 :: dlicol(0:1), dliexpco ! , expo_column

  real(kind=double), dimension(:),     pointer :: time_exp, time_raw, &
                                                  time_rnd
  integer(kind=Int8),  dimension(:),   pointer :: ccdnr
  integer(kind=Int32), dimension(:),   pointer :: outdlicol
  real(kind=single), dimension(:),     pointer :: dliexpcoCol 
  integer(kind=Int32), dimension(:,:,:)        :: dliarray(1:2,1:EPN_MAX_Y,0:12)


  integer(kind=Int16) :: quadrantId, &    ! quadrant identifier [0-3]
                         ccdId, &         ! CCD identifier      [0-2]
                         ccdno            ! overall CCD number  [1-12]
  integer, parameter  :: nccd = 12

  integer             :: i,j, irow, nrows, mrows, npar, dliodf_tot_quad(0:3), &
                         ic, ic_b(1:nccd), ic_e(1:nccd), &
                         trebin, ccdmode, col_per_quad, &
                         dliodf_total, dlisas_total, dli_total, &
                         ndsclinq, ndsclinq_quad(0:3), diff, an_chop
  real(kind=double)   :: an_dslin, dslin_pmh(0:3)
  real(kind=double)   :: timedelta, frameTime
  real(kind=double)   :: t_raw, t_exp, t_rnd, rr, ontime, nframe

  character(len=1024),dimension(:) :: eventsets(0:2) ! , &
!                                     framesets(0:2), colnames(0:13), &
!                                     coltypes(0:13), attribute(0:11) 
! character(len=1024) :: imagingset, mergedname, &
!                        auxfileset, phkfileset, auxfilename, phkfilename, &
!                        expression, filteredname
  character(len=8)    :: extname(0:nccd), colname(0:4), ext_events
  character(len=30)   :: subMode

  logical(kind=bool)  :: hasExtension
  logical             :: single_ccd_mode, done, lccd(0:nccd), &
                         dlimap_overflow = .FALSE., fastmode = .FALSE.

  logical             :: trandom   ! trigger for randomization of time within frame
  logical             :: escreen   ! trigger for removal of extra EXPOSU columns
  logical             :: espatial  ! trigger for determination of spatial exposure inhomogeneities (CCD columns)
  logical             :: lrebin    ! trigger for rebinning of spatial exposure inhomogeneities (CCD columns)
!
! compile time parameters
!
  logical, parameter  :: tdefault = .FALSE. ! trigger for backward compatibility:
                                   ! if T then TIME has unrandomized times (as always in the past) and TIME_RND is randomized
                                   ! if F then TIME_RAW has unrandomized times and TIME is randomized
  logical, parameter  :: use_expo = .FALSE. ! trigger to use EXPOSU## extension in time randomization
!
!
  write(errstr,*) 'EPEXPOSURE ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  single_ccd_mode = .TRUE.  
  quadrantId = 0
  CCDId = 0
  ccdmode = 0

!-------------------
! get all parameters  
!-------------------

  npar = parameterCount("eventsets")
  write(errstr,*) 'Parameter eventsets has ',npar,' entries'
  call message(AppMsg,Noisy)
  eventsets(0)=stringParameter("eventsets",0)
  write(errstr,*) 'Accessing ',trim(eventsets(0))
  call message(AppMsg,Noisy)

! trebin = integerParameter("rebinexposure")
  trebin = 1
  lrebin = (trebin > 1)
  if (lrebin) then
    call warning("rebinExposure","rebinning of EXPOSURE extension not yet enabled")
    lrebin = .FALSE.
  endif

  trandom = booleanParameter("randomizetime")
  if (trandom) then
!   call warning("randomizeTime","randomization of TIME not yet enabled")
!   trandom = .FALSE.
  endif

  espatial = booleanParameter("spatialexposure")
  if (espatial) then
    call warning("spatialExposure","determination of spatial (CCD column) exposure correction not yet enabled, dummy only.")
!   espatial = .FALSE.
  endif

  escreen = booleanParameter("screenexposure")

  done = (.NOT.(espatial .OR. trandom .OR. escreen))

  if (.NOT. done) then

    infile = dataSet(trim(eventsets(0)),MODIFY,USE_ENVIRONMENT)
!   infile = dataSet(trim(eventsets(0)),MODIFY,LOW_MEMORY)
!   infile = dataSet(trim(eventsets(0)),MODIFY,HIGH_MEMORY)

!-------------------------------------------------
! define EXPOSURE related extension names
!-------------------------------------------------
    extname(0) =  "EXPOSURE"
    do CCDno = 1,nccd
      write(extname(CCDno),'(a,i2.2)') "EXPOSU" ,CCDno
    enddo
    colname = (/'FRAMELIM','NDSCLIN1','NDSCLIN2','NDSCLIN3','NDSCLIN4'/)

!-------------------
! delete extra columns  
!-------------------
    if (escreen) then

      do CCDno = 0,nccd
        write(errstr,*) 'checking existence of ', trim(extname(CCDno))
        call message(AppMsg,Verbose)
        hasExtension = hasBlock(infile,trim(extname(CCDno)))
        if (hasExtension) then
          expo = table(infile,trim(extname(CCDno)))
          nrows = numberOfRows(expo)
          write(errstr,*) trim(extname(CCDno)),' has ',nrows,' number of rows'
          call message(AppMsg,Noisy)
          do i = 0,4
            if (hascolumn(expo,trim(colname(i)))) then
              write(errstr,*) trim(colname(I)),' will be removed now'
              call message(AppMsg,Noisy)
!             expo_column = column(expo,trim(colname(i)))
              call deleteColumn(expo,trim(colname(i)))
            endif
          enddo
          call release(expo)
        endif
      enddo

    endif

!-------------------------------------------------
! define DLIMAP related extension names
!-------------------------------------------------
    extname(0) =  "DLIMAP"
    do CCDno = 1,nccd
      write(extname(CCDno),'(a,i2.2)') "DLIMAP" ,CCDno
    enddo
    colname(0:1) = (/'DLIODF','DLISAS'/)
!
    lccd(0:nccd) = .false.
    do CCDno = 0,nccd
      write(errstr,*) 'checking existence of ', trim(extname(CCDno))
      call message(AppMsg,Verbose)
      hasExtension = hasBlock(infile,trim(extname(CCDno)))
      if (hasExtension) then
        lccd(CCDno) = .true.
        dlimap = table(infile,trim(extname(CCDno)))
        nrows = numberOfRows(dlimap)
        write(errstr,*) trim(extname(CCDno)),' has ',nrows,' number of rows'
        call message(AppMsg,Noisy)
        quadrantId = (CCDno-1)/3
        do i = 0,1
          if (hascolumn(dlimap,trim(colname(i)))) then
            write(errstr,*) trim(colname(I)),' will be analysed now'
            call message(AppMsg,Noisy)
            dlicol(i) = column(dlimap,trim(colname(i)))
            outdliCol => int32Data(dlicol(i))
            do j = 0,EPN_MAX_Y-1
              dliarray(i+1,j+1,CCDNo) = outdlicol(j)
            enddo
          endif
        enddo
        if (hasAttribute(dlimap,"CCDMODE")) ccdmode = int16Attribute(dlimap,"CCDMODE")
        call release(dlimap)
      else
        dliarray(1:2,1:EPN_MAX_Y,CCDNo) = 0
      endif
    enddo

    if (espatial) then
!
      ndsclinq_quad(0:3) = 0
      dliodf_tot_quad(0:3) = 0
      if (ccdmode .ge. 3) then
        single_CCD_mode = .true.
        col_per_quad = 64
      else
        col_per_quad = 192      
      endif
!
! DLI files show I*2 overflow if exposure is long
! proposed alogorithm needs to be recalibrated/redesigned
!
! append DLIEXPCO column of DLI-caused exposure correction factors
!
! works only for COLUMN-WISE MIP rejection (i.e. default)
!
! we need the following information:
!  ONTIME
!  FRAMETIM  -> these two give the number of frames
!  NDSCLINC  -> total number SAS per CCD
!  NDSCLINQ  -> total number onboard per quadrant
!  AN_DSLIN  -> with this number one can estimate the probability of
!               I*2 overflow in the DLIMAP, per 20 frames per quadrant
!
      do CCDno = 0,nccd
        quadrantId = (CCDno-1)/3
        if (lccd(CCDno)) then
          if (CCDno .eq. 0) single_CCD_mode = .true.
          write(errstr,*) 'updating ', trim(extname(CCDno))
          call message(AppMsg,Verbose)
          dlimap = table(infile,trim(extname(CCDno)))
!
          if (hasAttribute(dlimap,"AN_CHOP")) then
            an_chop = int16Attribute(dlimap,"AN_CHOP")
          else
            an_chop = 0 ! no chopper
          endif
!
          if (hasAttribute(dlimap,"FRAMETIM")) then
            frameTime = real32Attribute(dlimap,"FRAMETIM")
          else
            frameTime = 73.36 ! ms FF
          endif
!
          if (hasAttribute(dlimap,"ONTIME")) then
            ontime = real32Attribute(dlimap,"ONTIME")
          else
            ontime = 1.0 ! s
          endif
!
          if (hasAttribute(dlimap,"AN_DSLIN")) then
            an_dslin = int32Attribute(dlimap,"AN_DSLIN")
          else
            an_dslin = 140.0 ! per 20 frames per quadrant, from PNPMH1
          endif
      ! is this still true for chopper ?
          dslin_pmh(quadrantId) = ontime/frametime*1.d3/(20)*an_dslin
!
          if (hasAttribute(dlimap,"NDSCLINQ")) then
            ndsclinq = int32Attribute(dlimap,"NDSCLINQ")
          else
            ndsclinq = 1 ! per exposure per quadrant, from PNAUX2
          endif
          dlimap_overflow = (ndsclinq/(1.0*col_per_quad) .gt. 64000)
          if (dlimap_overflow) then
            call warning("spatialDLI", &
              "discarded line files may be affected by I*2 overflow.")
          endif
!
          nframe = NINT(ontime/frameTime*1.d3)
!
          if (hasColumn(dlimap,"DLIEXPCO")) then
            dliexpco = column(dlimap,"DLIEXPCO",MODIFY)  
          else
            dliexpco = addColumn(dlimap,"DLIEXPCO",Real32,"")
          endif
          dliexpcoCol => real32Data(dliexpco)
            
!         if (single_CCD_mode) then
            dliexpcoCol(0:EPN_MAX_Y-1) = 0.0
            dli_total = 0
            dliodf_total = 0
            dlisas_total = 0
            do j = 0,EPN_MAX_X-1  ! [1..64]
              dliodf_total = dliodf_total + dliarray(1,j+1,CCDno)
              dlisas_total = dlisas_total + dliarray(2,j+1,CCDno)
            enddo
            if (ccdno .gt. 0) then
              dliodf_tot_quad(quadrantId) = dliodf_tot_quad(quadrantId)+dliodf_total
              ndsclinq_quad(quadrantId) = ndsclinq
            endif
!
! issue a warning if DLIODF columns per quadrant are significantly
! lower than the NDSCLINQ columns (not affected by I*2 overflow!)
!
            dli_total = dliodf_total+dlisas_total
            do j = 0,EPN_MAX_X-1
              dliexpcoCol(j) = 1.0 ! (1.0d0*EPN_MAX_X)/total*dliarray(i+1,j+1,CCDno)
            enddo
!         else
!           dliexpcoCol(0:EPN_MAX_Y-1) = 1.0
!         endif
          call release(dliexpco)
          call release(dlimap)
        endif
      enddo
      do quadrantId = 0,3
        diff = dliodf_tot_quad(quadrantId)-ndsclinq_quad(quadrantId)
        write(errstr,*) 'Quadrant ',quadrantID,':  PMH:', &
          NINT(dslin_pmh(quadrantId)),'  PNAUX:', &
          ndsclinq_quad(quadrantId),'  DLI:',dliodf_tot_quad(quadrantId), &
          '  DIFF:',diff
        call message(AppMsg,Verbose)
        dlimap_overflow = (diff .lt. -15000)
        if (dlimap_overflow) then
          write(errstr,*) 'discarded line files may be affected by I*2 overflow in quadrant',quadrantID
          call warning('spatialDLI', trim(errstr))
          write(errstr,'(1x,a,i1,a,i5,a)') &
               'The PNAUX-DLI difference in Q',quadrantId,' corresponds to about',&
                NINT(ABS(1.0*diff)/65536),' I*2 overflows'
          call message(AppMsg,Sparse)
        endif
        dlimap_overflow = (diff .gt. 20000)
        if (dlimap_overflow) then
          write(errstr,*) 'discarded line files show significant deficiency (low quality) in quadrant',quadrantID
          call warning('spatialDLI', trim(errstr))
          write(errstr,'(1x,a,i1,a,i5,a)') &
               'The DLI-PNAUX difference in Q',quadrantId,' corresponds to about',&
                NINT(ABS(1.0*diff)/65536),' I*2 overflows'
          call message(AppMsg,Sparse)
        endif
      enddo
    endif

!-----------------------------------------------------------------------
! do randomization in TIME within one readout frame
! NOTE: it is assumed that the files are fractionally ordered,
!   i.e. for each (sorted) CCDNR value the events are sorted in TIME;
!   this is the case for normal EPIC-pn SAS event files (PPS, epchain)
!-----------------------------------------------------------------------

    extname(0) =  "EXPOSURE"
    do CCDno = 1,nccd
      write(extname(CCDno),'(a,i2.2)') "EXPOSU" ,CCDno
    enddo
    colname = (/'FRAMELIM','NDSCLIN1','NDSCLIN2','NDSCLIN3','NDSCLIN4'/)

    do CCDno = 0,nccd
      hasExtension = hasBlock(infile,trim(extname(CCDno)))
      if (hasExtension) then
        expo = table(infile,trim(extname(CCDno)))
        call setAttribute(expo, "RAND_TIM", 0, "", "1: randomized time 0: not randomized")
        call release(expo)
      end if
    end do  

    if (trandom) then
      ext_events = 'EVENTS'
      write(errstr,*) 'checking existence of ', trim(ext_events)
      call message(AppMsg,Verbose)
      hasExtension = hasBlock(infile,trim(ext_events))
      if (hasExtension) then
        events = table(infile,trim(ext_events))
        if (hasAttribute(infile,"SUBMODE")) then
          submode = trim(stringAttribute(infile,"SUBMODE"))
          write(errstr,'(1x,a,t39,a)') 'SubMode:     ',trim(submode)
          call message(AppMsg,Verbose)
!
! no recipe yet for FAST modes (TI+BU), so issue warning and do nothing here:
!
          if (trim(submode) .eq. trim(modeId(FAST_TIMING)) .or. &
              trim(submode) .eq. trim(modeId(FAST_BURST))  ) then
            fastmode = .TRUE.
            call warning("randomizeTime", &
              "randomization of TIME not yet enabled for FAST modes (TI,BU).")
          else
            fastmode = .FALSE.
          endif
        else
          submode = ' '
          call message(AppMsg,Verbose,'SUBMODE attribute not found')
        endif
        if (hascolumn(events,'TIME').and.hascolumn(events,'CCDNR').and.(.NOT.fastmode)) then
          nrows = numberOfRows(events)
          if (nrows .eq. 0) then
            ! empty extension !!
          endif
          if (tdefault) then
            if (hascolumn(events,'TIME_RND')) then
!             nothing to be done, already backup column present
            else
              call copyColumn(events,column(events,'TIME'),'TIME_RND')
            endif
          else
            if (hascolumn(events,'TIME_RAW')) then
!             nothing to be done, already backup column present
            else
              call copyColumn(events,column(events,'TIME'),'TIME_RAW')
            endif
          endif
!
! always use the original TIME column for randomization as it already
! contains the proper SAS fine time corrections
!
          if (tdefault) then
            time_raw  => real64Data(column(events,'TIME'))
            time_rnd  => real64Data(column(events,'TIME_RND'))
          else
            time_raw  => real64Data(column(events,'TIME_RAW'))
            time_rnd  => real64Data(column(events,'TIME'))
          endif
          ccdnr => int8Data(column(events,'CCDNR'))
!
! read start time (first event), however, not used later at the moment
!
          t_raw = time_raw(0)
          t_rnd = time_rnd(0)
!
! check which blocks of events belong to a certain CCD
!
          lccd(1:nccd) = .false.
          ic_b(1:nccd) = 0
          ic_e(1:nccd) = 0
!
          ic = ccdnr(0)
          ic_b(ic) = 1
          lccd(ic) = .true.
          do irow = 1, nrows-1
            if (ccdnr(irow).gt.ic) then
              ic_e(ic) = irow-2
              ic = ccdnr(irow)
              ic_b(ic) = irow-1
              lccd(ic) = .true.
            endif
          enddo
          ic_e(ic) = nrows
          do ic = 1,nccd
            write(errstr,'(a,i2,a,i8,a,i8)') 'CCD ',ic,': from event ',&
                ic_b(ic),' to ',ic_e(ic)
            call message(AppMsg,Noisy)
          enddo
!         nullify(ccdnr)

          do CCDno = 0,nccd
            if (CCDno.ge.1) then
              if (.NOT.lccd(ccdno)) cycle ! no events with this CCD number
            endif
            write(errstr,*) 'checking existence of ', trim(extname(CCDno))
            call message(AppMsg,Verbose)
            hasExtension = hasBlock(infile,trim(extname(CCDno)))
            if (hasExtension) then
              expo = table(infile,trim(extname(CCDno)))
              mrows = numberOfRows(expo)
              write(errstr,*) trim(extname(CCDno)),' has ',mrows,' number of rows'
              call message(AppMsg,Noisy)
!
! TIMEDEL is *not* the proper quantity for randomization at it gives only
! the exposure time under the PSF but does not include out-of-time events,
! it has to be replaced by the total frame time minus real dead time (later)
!
              if (hasAttribute(expo,"TIMEDEL")) then
                timedelta = real32Attribute(expo,"TIMEDEL")
                write(errstr,'(1x,a,t39,f10.6,a)') 'TimeDel:',timedelta,' [s]'
                call message(AppMsg,Verbose)
               else
                submode = ' '
                call message(AppMsg,Verbose,'TIMEDEL attribute not found')
                timedelta = 0.0d0
              endif
              if (use_expo) then
                time_exp => real64Data(column(expo,'TIME'))
                t_exp = time_exp(0)
              endif
              if (CCDno.eq.0) then
                do irow = 0,nrows-1
                  call random_number(rr)
                  time_rnd(irow) = time_raw(irow)+(rr-0.5d0)*timedelta
                enddo
              else
                write(errstr,'(1x,a,t39,2i9)') 'CCD index range:',ic_b(CCDno),ic_e(CCDno)
                call message(AppMsg,Noisy)
                do irow = ic_b(CCDno)-1,ic_e(CCDno)-1
                  call random_number(rr)
                  time_rnd(irow) = time_raw(irow)+(rr-0.5d0)*timedelta
                enddo
              endif
              call setAttribute(expo, "RAND_TIM", 1, "", "1: randomized time 0: not randomized")
              call release(expo)
            endif
          enddo
!         nullify(time_exp)
!         nullify(time_rnd)
!         nullify(time_raw)
        else ! does not have TIME and/or CCDNR columns

        endif
        call release(events)

      else ! does not have EVENTS

      endif

    endif

!-------------------------------------------------
! write out modified event file
!-------------------------------------------------
    call message(AppMsg,Verbose,'Writing output file')
    call release(infile)
  
  else
    call message(AppMsg,Sparse,'Nothing to be done for input file(s).')
  endif


  end subroutine epexposure

end module epexposure_mod
